@@ -1,7 +1,7 @@
 # Definition of the processing context for Wikitext processing, and code for
 # expanding templates, parser functions, and Lua macros.
 #
-# Copyright (c) 2020 Tatu Ylonen.  See file LICENSE and https://ylonen.org
+# Copyright (c) 2020-2021 Tatu Ylonen.  See file LICENSE and https://ylonen.org
 
 import os
 import re
@@ -102,10 +102,10 @@ class Wtp(object):
         "subsection",    # Subsection within page, for error messages
         "suppress_special",  # XXX never set to True???
     )
-    def __init__(self, quiet=False, num_threads=None, cache_file=None):
-        assert quiet in (True, False)
+    def __init__(self, num_threads=None, cache_file=None, quiet=False):
         assert num_threads is None or isinstance(num_threads, int)
         assert cache_file is None or isinstance(cache_file, str)
+        assert quiet in (True, False)
         if num_threads is None:
             import platform
             if platform.system() in ("Windows", "Darwin"):
@@ -1009,8 +1009,9 @@ class Wtp(object):
 
                     # Check for undefined templates
                     if name not in all_templates:
-                        self.warning("undefined template {!r}"
-                                     .format(tname))
+                        # XXX tons of these in enwiktionary-20201201 ???
+                        #self.warning("undefined template {!r}"
+                        #             .format(tname))
                         parts.append('<strong class="error">Template:{}'
                                      '</strong>'
                                      .format(html.escape(name)))
@@ -1037,7 +1038,8 @@ class Wtp(object):
                     num = 1
                     for i in range(1, len(args)):
                         arg = str(args[i])
-                        m = re.match(r"""^\s*([^<>="']+?)\s*=\s*(.*?)\s*$""",
+                        m = re.match(r"""(?s)^\s*([^<>="']+?)\s*="""
+                                     """\s*(.*?)\s*$""",
                                      arg)
                         if m:
                             # Note: Whitespace is stripped by the regexp
@@ -1076,7 +1078,7 @@ class Wtp(object):
                         t = template_fn(urllib.parse.unquote(name), ht)
                     if t is None:
                         if name in self.transient_templates:
-                            body = self.transient_templates[body]
+                            body = self.transient_templates[name]
                         else:
                             body = self.templates[name]
                         # XXX optimize by pre-encoding bodies during
@@ -1228,6 +1230,8 @@ class Wtp(object):
         call page_handler in parallel, and thus page_handler should
         not attempt to save anything between calls and should not
         modify global data.  This function is not re-entrant."""
+        assert callable(page_handler)
+        assert autoload in (True, False)
         global _global_ctx
         global _global_page_handler
         global _global_page_autoload
@@ -1254,18 +1258,18 @@ class Wtp(object):
                 pool = multiprocessing.Pool(self.num_threads)
             cnt = 0
             for success, ret in pool.imap_unordered(phase2_page_handler,
-                                                    self.page_seq):
+                                                    self.page_seq, 64):
                 if not success:
                     print(ret)  # Print error in parent process - do not remove
                     continue
                 if ret is not None:
                     yield ret
-                    cnt += 1
-                    if not self.quiet and cnt % 1000 == 0:
-                        print("  ... {}/{} pages ({:.1%}) processed"
-                              .format(cnt, len(self.page_seq),
-                                      cnt / len(self.page_seq)))
-                        sys.stdout.flush()
+                cnt += 1
+                if not self.quiet and cnt % 1000 == 0:
+                    print("  ... {}/{} pages ({:.1%}) processed"
+                          .format(cnt, len(self.page_seq),
+                                  cnt / len(self.page_seq)))
+                    sys.stdout.flush()
             pool.close()
             pool.join()
 
