@@ -4,15 +4,11 @@
 
 import re
 import enum
-import html
 from .parserfns import PARSER_FUNCTIONS
 from .wikihtml import ALLOWED_HTML_TAGS
 from .common import MAGIC_NOWIKI_CHAR, MAGIC_FIRST, MAGIC_LAST
 
 
-# HTML tags that are also parsed in the preparse phase
-PRE_PARSE_TAGS = ["noinclude", "includeonly", "onlyinclude"]
-
 # Set of tags that can be parents of "flow" parents
 HTML_FLOW_PARENTS = set(k for k, v in ALLOWED_HTML_TAGS.items()
                         if "flow" in v.get("content", [])
@@ -38,6 +34,13 @@ HTML_PERMITTED_PARENTS = {
 }
 HTML_PERMITTED_PARENTS["text"] = HTML_PHRASING_PARENTS
 
+# Set of HTML tag like names that we treat as literal without any warning
+SILENT_HTML_LIKE = set([
+    "gu",
+    "qu",
+    "e",
+])
+
 
 # MediaWiki magic words.  See https://www.mediawiki.org/wiki/Help:Magic_words
 MAGIC_WORDS = set([
@@ -131,7 +134,7 @@ class NodeKind(enum.Enum):
     # In WikiText {{name|arg1|...}}.
     TEMPLATE = enum.auto(),
 
-    # A template argument expansion.  Variable name is in first argument and
+    # A template argument expansion.  Argument name is in first argument and
     # subsequent arguments in remaining arguments.  Children are not used.
     # In WikiText {{{name|...}}}
     TEMPLATE_ARG = enum.auto(),
@@ -170,17 +173,10 @@ class NodeKind(enum.Enum):
     # merged into a single node and the content between them is stored
     # in the node's children.  Args is the name of the tag directly
     # (i.e., not a list and always without a slash).  Attrs contains
-    # attributes from the HTML start tag.  The special tags
-    # <onlyinclude>, <noinclude>, and <includeonly> as well as WikiText-related
-    # tags with HTML-like syntax also generate this tag (with the exception
-    # of <pre> and <nowiki>, which are handled specially).
+    # attributes from the HTML start tag.  Contents in a paired tag
+    # are stored in ``children``.
     HTML = enum.auto(),
 
-    # XXX <ref ...> and <references />
-    # XXX -{ ... }- syntax, see
-
-    # XXX __NOTOC__
-
 
 # Maps subtitle token to its kind
 subtitle_to_kind = {
@@ -241,7 +237,7 @@ class WikiNode(object):
         self.args = []  # List of lists
         self.attrs = {}
         self.children = []   # list of str and WikiNode
-        self.loc = loc
+        self.loc = loc  # XXX is this used???
 
     def __str__(self):
         return "<{}({}){} {}>".format(self.kind.name,
@@ -259,39 +255,38 @@ def _parser_push(ctx, kind):
     assert isinstance(kind, NodeKind)
     _parser_merge_str_children(ctx)
     node = WikiNode(kind, ctx.linenum)
-    prev = ctx.stack[-1]
+    prev = ctx.parser_stack[-1]
     prev.children.append(node)
-    ctx.stack.append(node)
+    ctx.parser_stack.append(node)
     ctx.suppress_special = False
     return node
 
+
 def _parser_merge_str_children(ctx):
     """Merges multiple consecutive str children into one.  We merge them
     as a separate step, because this gives linear worst-case time, vs.
     quadratic worst case (albeit with lower constant factor) if we just
-    added to the previously accumulated string in text_fn() instead."""
-    node = ctx.stack[-1]
-    lst = node.children
-    lstlen = len(lst)
-    for i in range(lstlen - 1, -1, -1):
-        if not isinstance(lst[i], str):
-            break
-    else:
-        # All children are strings
-        s = "".join(lst)
-        s = re.sub(MAGIC_NOWIKI_CHAR, "", s)
-        node.children = []
+    added to the previously accumulated string in text_fn() instead.
+    Importantly, this also finalizes string children so that any magic
+    characters are expanded and nowiki characters removed."""
+    node = ctx.parser_stack[-1]
+    new_children = []
+    strings = []
+    for x in node.children:
+        if isinstance(x, str):
+            strings.append(x)
+        else:
+            if strings:
+                s = ctx._finalize_expand("".join(strings), False)
+                if s:
+                    new_children.append(s)
+                strings = []
+            new_children.append(x)
+    if strings:
+        s = ctx._finalize_expand("".join(strings), False)
         if s:
-            node.children.append(s)
-        return
-    cnt = lstlen - i - 1
-    if cnt == 0:
-        return
-    node.children = lst[:-cnt]
-    s = "".join(lst[-cnt:])
-    s = re.sub(MAGIC_NOWIKI_CHAR, "", s)
-    if s:
-        node.children.append(s)
+            new_children.append(s)
+    node.children = new_children
 
 def _parser_pop(ctx, warn_unclosed):
     """Pops a node from the stack.  If the node has arguments, this moves
@@ -302,28 +297,45 @@ def _parser_pop(ctx, warn_unclosed):
     the nodes when their parsing completes."""
     assert warn_unclosed in (True, False)
     _parser_merge_str_children(ctx)
-    node = ctx.stack[-1]
+    node = ctx.parser_stack[-1]
 
     # Warn about unclosed syntaxes.
     if warn_unclosed and node.kind in MUST_CLOSE_KINDS:
         if node.kind == NodeKind.HTML:
-            ctx.error("HTML tag <{}> not properly closed, started on "
-                       "line {}".format(node.args, node.loc))
+            ctx.warning("HTML tag <{}> not properly closed".format(node.args),
+                        trace="started on line {}, detected on line {}"
+                        .format(node.loc, ctx.linenum))
         elif node.kind == NodeKind.PARSER_FN:
-            ctx.error("parser function {!r} not properly closed, started on "
-                      "line {}".format(node.args[0], node.loc))
+            ctx.warning("parser function {!r} not properly closed"
+                        .format(node.args[0]),
+                        trace="started on line {}, detected on line {}"
+                        .format(node.loc, ctx.linenum))
+        elif node.kind == NodeKind.URL and not node.children:
+            # This can happen at least when [ is inside tempate argument.
+            ctx.parser_stack.pop()
+            node2 = ctx.parser_stack[-1]
+            node3 = node2.children.pop()
+            assert node3 is node
+            text_fn(ctx, "[")
+            return
+        elif node.kind in (NodeKind.ITALIC, NodeKind.BOLD):
+            # Unbalanced italic/bold annotation is so extremely common
+            # in Wiktionary that let's suppress any warnings about
+            # them.
+            pass
         else:
-            ctx.error("{} not properly closed, started on line {}"
-                       "".format(node.kind.name, node.loc))
+            ctx.warning("{} not properly closed".format(node.kind.name),
+                        trace="started on line {}, detected on line {}"
+                        .format(node.loc, ctx.linenum))
 
     # When popping BOLD and ITALIC nodes, if the node has no children,
     # just remove the node from it's parent's children.  We may otherwise
     # generate spurious empty BOLD and ITALIC nodes when closing them
     # out-of-order (which happens always with '''''bolditalic''''').
     if node.kind in (NodeKind.BOLD, NodeKind.ITALIC) and not node.children:
-        ctx.stack.pop()
-        assert ctx.stack[-1].children[-1].kind == node.kind
-        ctx.stack[-1].children.pop()
+        ctx.parser_stack.pop()
+        assert ctx.parser_stack[-1].children[-1].kind == node.kind
+        ctx.parser_stack[-1].children.pop()
         return
 
     # If the node has arguments, move remamining children to be the last
@@ -354,12 +366,12 @@ def _parser_pop(ctx, warn_unclosed):
 
     # Remove the topmost node from the stack.  It should be on its parent's
     # chilren list.
-    ctx.stack.pop()
+    ctx.parser_stack.pop()
 
 def _parser_have(ctx, kind):
     """Returns True if any node on the stack is of the given kind."""
     assert isinstance(kind, NodeKind)
-    for node in ctx.stack:
+    for node in ctx.parser_stack:
         if node.kind == kind:
             return True
     return False
@@ -367,27 +379,44 @@ def _parser_have(ctx, kind):
 
 def text_fn(ctx, token):
     """Inserts the token as raw text into the parse tree."""
-    node = ctx.stack[-1]
+    node = ctx.parser_stack[-1]
 
-    # Whitespaces inside an external link divide its first argument from its
-    # second argument.  All remaining words go into the second argument.
-    if token.isspace() and node.kind == NodeKind.URL and not node.args:
-        _parser_merge_str_children(ctx)
-        node.args.append(node.children)
-        node.children = []
-        return
+    # Convert certain characters from the token into HTML entities
+    token = re.sub(r"<", "&lt;", token)
+    token = re.sub(r">", "&gt;", token)
+
+    # External links [https://...] require some magic.  They only seem to
+    # be links if the content looks like a URL."""
+    if node.kind == NodeKind.URL:
+        if not node.args and not node.children:
+            if not re.match(r"^(http|https|mailto):", token):
+                # It does not look like a URL
+                ctx.parser_stack.pop()
+                node2 = ctx.parser_stack[-1]
+                node3 = node2.children.pop()
+                assert node3 is node
+                return text_fn(ctx, "[" + token)
+
+        # Whitespaces inside an external link divide its first argument from its
+        # second argument.  All remaining words go into the second argument.
+        if token.isspace() and not node.args:
+            _parser_merge_str_children(ctx)
+            node.args.append(node.children)
+            node.children = []
+            return
 
     # Some nodes are automatically popped on newline/text
     if ctx.beginning_of_line:
         while True:
-            node = ctx.stack[-1]
+            node = ctx.parser_stack[-1]
             if node.kind == NodeKind.LIST_ITEM:
                 if token.startswith(" ") or token[0].startswith("\t"):
                     node.children.append(token)
                     return
                 _parser_merge_str_children(ctx)
                 if (node.children and isinstance(node.children[-1], str) and
-                    not node.children[-1].isspace() and
+                    (len(node.children) > 1 or
+                     not node.children[-1].isspace()) and
                     node.children[-1].endswith("\n")):
                     _parser_pop(ctx, False)
                     continue
@@ -401,11 +430,16 @@ def text_fn(ctx, token):
                     not token.startswith(" ") and not token.isspace()):
                     _parser_pop(ctx, False)
                     continue
+            elif node.kind in (NodeKind.BOLD, NodeKind.ITALIC):
+                _parser_merge_str_children(ctx)
+                ctx.warning("{} not properly closed on the same line"
+                            .format(node.kind.name))
+                _parser_pop(ctx, False)
             break
 
         # Spaces at the beginning of a line indicate preformatted text
         if token.startswith(" ") or token.startswith("\t"):
-            if node.kind != NodeKind.PREFORMATTED:
+            if node.kind != NodeKind.PREFORMATTED and not ctx.pre_parse:
                 node = _parser_push(ctx, NodeKind.PREFORMATTED)
 
     # If the previous child was a link that doesn't yet have children,
@@ -432,7 +466,7 @@ def hline_fn(ctx, token):
     # table element.  We also won't pop HTML nodes as they might appear
     # in template definitions.
     while True:
-        node = ctx.stack[-1]
+        node = ctx.parser_stack[-1]
         if node.kind in (NodeKind.ROOT, NodeKind.LEVEL2,
                          NodeKind.TABLE, NodeKind.TABLE_CAPTION,
                          NodeKind.TABLE_ROW, NodeKind.TABLE_HEADER_CELL,
@@ -457,8 +491,8 @@ def subtitle_start_fn(ctx, token):
     # is of a higher level - but only if there are remaining subtitles.
     # Subtitles sometimes occur inside <noinclude> and similar tags, and we
     # don't want to force closing those.
-    while any(x.kind in kind_to_level for x in ctx.stack):
-        node = ctx.stack[-1]
+    while any(x.kind in kind_to_level for x in ctx.parser_stack):
+        node = ctx.parser_stack[-1]
         if kind_to_level.get(node.kind, 99) < level:
             break
         if node.kind == NodeKind.HTML:
@@ -480,15 +514,15 @@ def subtitle_end_fn(ctx, token):
 
     # Keep popping formats until we get to the subtitle node
     while True:
-        node = ctx.stack[-1]
+        node = ctx.parser_stack[-1]
         if node.kind in kind_to_level:
             break
         _parser_pop(ctx, True)
 
     # Move children of the subtitle node to be its first argument.
-    node = ctx.stack[-1]
+    node = ctx.parser_stack[-1]
     if node.kind != kind:
-        ctx.error("subtitle start and end markers level mismatch")
+        ctx.warning("subtitle start and end markers level mismatch")
     _parser_merge_str_children(ctx)
     node.args.append(node.children)
     node.children = []
@@ -499,7 +533,13 @@ def italic_fn(ctx, token):
     if ctx.pre_parse:
         return text_fn(ctx, token)
 
-    if not _parser_have(ctx, NodeKind.ITALIC):
+    node = ctx.parser_stack[-1]
+
+    if node.kind in (NodeKind.TEMPLATE, NodeKind.TEMPLATE_ARG):
+        return text_fn(ctx, token)
+
+    if (not _parser_have(ctx, NodeKind.ITALIC) or
+        node.kind in (NodeKind.LINK,)):
         # Push new formatting node
         _parser_push(ctx, NodeKind.ITALIC)
         return
@@ -508,7 +548,7 @@ def italic_fn(ctx, token):
     # to allow closing them in either order.
     push_bold = False
     while True:
-        node = ctx.stack[-1]
+        node = ctx.parser_stack[-1]
         if node.kind == NodeKind.ITALIC:
             _parser_pop(ctx, False)
             break
@@ -524,7 +564,13 @@ def bold_fn(ctx, token):
     if ctx.pre_parse:
         return text_fn(ctx, token)
 
-    if not _parser_have(ctx, NodeKind.BOLD):
+    node = ctx.parser_stack[-1]
+
+    if node.kind in (NodeKind.TEMPLATE, NodeKind.TEMPLATE_ARG):
+        return text_fn(ctx, token)
+
+    if (not _parser_have(ctx, NodeKind.BOLD) or
+        node.kind in (NodeKind.LINK,)):
         # Push new formatting node
         _parser_push(ctx, NodeKind.BOLD)
         return
@@ -533,7 +579,7 @@ def bold_fn(ctx, token):
     # to allow closing them in either order.
     push_italic = False
     while True:
-        node = ctx.stack[-1]
+        node = ctx.parser_stack[-1]
         if node.kind == NodeKind.BOLD:
             _parser_pop(ctx, False)
             break
@@ -568,7 +614,7 @@ def elink_end_fn(ctx, token):
     if not _parser_have(ctx, NodeKind.URL):
         return text_fn(ctx, token)
     while True:
-        node = ctx.stack[-1]
+        node = ctx.parser_stack[-1]
         if node.kind == NodeKind.URL:
             _parser_pop(ctx, False)
             break
@@ -581,12 +627,21 @@ def url_fn(ctx, token):
     if ctx.pre_parse:
         return text_fn(ctx, token)
 
-    node = ctx.stack[-1]
+    # If the URL ends in certain common punctuation characters, put the
+    # punctuation as text after it.
+    suffix = None
+    if token[-1] in ".!?,":
+        suffix = token[-1]
+        token = token[:-1]
+
+    node = ctx.parser_stack[-1]
     if node.kind == NodeKind.URL:
         return text_fn(ctx, token)
     node = _parser_push(ctx, NodeKind.URL)
     text_fn(ctx, token)
     _parser_pop(ctx, False)
+    if suffix:
+        text_fn(ctx, suffix)
 
 
 def magic_fn(ctx, token):
@@ -611,7 +666,7 @@ def magic_fn(ctx, token):
             process_text(ctx, arg)
 
         while True:
-            node = ctx.stack[-1]
+            node = ctx.parser_stack[-1]
             if node.kind == NodeKind.ROOT:
                 break
             if node.kind in (NodeKind.TEMPLATE, NodeKind.PARSER_FN):
@@ -635,7 +690,7 @@ def magic_fn(ctx, token):
             process_text(ctx, arg)
 
         while True:
-            node = ctx.stack[-1]
+            node = ctx.parser_stack[-1]
             if node.kind == NodeKind.ROOT:
                 break
             if node.kind == NodeKind.TEMPLATE_ARG:
@@ -644,19 +699,20 @@ def magic_fn(ctx, token):
             _parser_pop(ctx, True)
 
     elif kind == "L":
-        assert len(args) == 1
         if nowiki:
             process_text(ctx, "&lsqb;&lsqb;" + args[0] + "&rsqb;&rsqb;")
             return
         # Link to another page
         _parser_push(ctx, NodeKind.LINK)
 
-        # Process the only argument
-        assert len(args) == 1
+        # Process arguments
         process_text(ctx, args[0])
+        for arg in args[1:]:
+            vbar_fn(ctx, "|")
+            process_text(ctx, arg)
 
         while True:
-            node = ctx.stack[-1]
+            node = ctx.parser_stack[-1]
             if node.kind == NodeKind.ROOT:
                 break
             if node.kind == NodeKind.LINK:
@@ -672,7 +728,7 @@ def colon_fn(ctx, token):
     """Handler for a special colon ":" within a template call.  This indicates
     that it is actually a parser function call.  This is called from list_fn()
     when it detects that it is inside a template node."""
-    node = ctx.stack[-1]
+    node = ctx.parser_stack[-1]
 
     # Unless we are in the first argument of a template, treat a colon that is
     # not at the beginning of a
@@ -706,7 +762,7 @@ def table_start_fn(ctx, token):
 
 def table_check_attrs(ctx):
     """Checks if the table has attributes, and if so, parses them."""
-    node = ctx.stack[-1]
+    node = ctx.parser_stack[-1]
     if node.kind != NodeKind.TABLE:
         return
     _parser_merge_str_children(ctx)
@@ -718,7 +774,7 @@ def table_check_attrs(ctx):
 
 def table_row_check_attrs(ctx):
     """Checks if the table row has attributes, and if so, parses them."""
-    node = ctx.stack[-1]
+    node = ctx.parser_stack[-1]
     if node.kind != NodeKind.TABLE_ROW:
         return
     _parser_merge_str_children(ctx)
@@ -737,7 +793,7 @@ def table_caption_fn(ctx, token):
     if not _parser_have(ctx, NodeKind.TABLE):
         return text_fn(ctx, token)
     while True:
-        node = ctx.stack[-1]
+        node = ctx.parser_stack[-1]
         if node.kind == NodeKind.TABLE:
             break
         _parser_pop(ctx, True)
@@ -749,12 +805,15 @@ def table_hdr_cell_fn(ctx, token):
     if ctx.pre_parse:
         return text_fn(ctx, token)
 
+    if token == "!" and not ctx.beginning_of_line:
+        return text_fn(ctx, token)
+
     table_row_check_attrs(ctx)
     table_check_attrs(ctx)
     if not _parser_have(ctx, NodeKind.TABLE):
         return text_fn(ctx, token)
     while True:
-        node = ctx.stack[-1]
+        node = ctx.parser_stack[-1]
         if node.kind == NodeKind.TABLE_ROW:
             _parser_push(ctx, NodeKind.TABLE_HEADER_CELL)
             return
@@ -784,7 +843,7 @@ def table_row_fn(ctx, token):
     if not _parser_have(ctx, NodeKind.TABLE):
         return text_fn(ctx, token)
     while True:
-        node = ctx.stack[-1]
+        node = ctx.parser_stack[-1]
         if node.kind == NodeKind.TABLE:
             break
         _parser_pop(ctx, True)
@@ -806,7 +865,7 @@ def table_cell_fn(ctx, token):
         # This might separate attributes for captions, header cells, and
         # data cells
         _parser_merge_str_children(ctx)
-        node = ctx.stack[-1]
+        node = ctx.parser_stack[-1]
         if (not node.attrs and len(node.children) == 1 and
             isinstance(node.children[0], str)):
             if node.kind in (NodeKind.TABLE_CAPTION,
@@ -818,7 +877,7 @@ def table_cell_fn(ctx, token):
         return text_fn(ctx, token)
 
     while True:
-        node = ctx.stack[-1]
+        node = ctx.parser_stack[-1]
         if node.kind == NodeKind.TABLE_ROW:
             break
         if node.kind == NodeKind.TABLE:
@@ -835,7 +894,7 @@ def vbar_fn(ctx, token):
     the vertical bar depends on context; it can separate arguments to
     templates, template argument references, links, etc, and it can
     also separate table row cells."""
-    node = ctx.stack[-1]
+    node = ctx.parser_stack[-1]
     if node.kind in HAVE_ARGS_KINDS:
         _parser_merge_str_children(ctx)
         node.args.append(node.children)
@@ -849,7 +908,7 @@ def double_vbar_fn(ctx, token):
     """Handle function for double vertical bar ||.  This is used as a column
     separator in tables.  If it occurs in other contexts, it should be
     interpreted as two vertical bars."""
-    node = ctx.stack[-1]
+    node = ctx.parser_stack[-1]
     if node.kind in HAVE_ARGS_KINDS:
         vbar_fn(ctx, "|")
         vbar_fn(ctx, "|")
@@ -863,10 +922,12 @@ def table_end_fn(ctx, token):
     if ctx.pre_parse:
         return text_fn(ctx, token)
 
+    table_row_check_attrs(ctx)
+    table_check_attrs(ctx)
     if not _parser_have(ctx, NodeKind.TABLE):
         return text_fn(ctx, token)
     while True:
-        node = ctx.stack[-1]
+        node = ctx.parser_stack[-1]
         if node.kind == NodeKind.TABLE:
             _parser_pop(ctx, False)
             break
@@ -880,7 +941,7 @@ def list_fn(ctx, token):
     if ctx.pre_parse:
         return text_fn(ctx, token)
 
-    node = ctx.stack[-1]
+    node = ctx.parser_stack[-1]
 
     # A colon inside a template means it is a parser function call.  We use
     # colon_fn() to handle that kind of colon.
@@ -897,7 +958,7 @@ def list_fn(ctx, token):
     # in the middle of a line.  Some of these cases were handled above; some
     # are handled here.
     if not ctx.beginning_of_line:
-        node = ctx.stack[-1]
+        node = ctx.parser_stack[-1]
         if (token == ":" and node.kind == NodeKind.LIST_ITEM and
             node.args.endswith(";") and "head" not in node.attrs):
             # Got definition for a head in a definition list on the same line
@@ -912,7 +973,7 @@ def list_fn(ctx, token):
 
     # Pop any lower-level list items
     while True:
-        node = ctx.stack[-1]
+        node = ctx.parser_stack[-1]
 
         # Check for a definition in a definition list
         if (node.kind == NodeKind.LIST_ITEM and node.args.endswith(";") and
@@ -929,7 +990,8 @@ def list_fn(ctx, token):
         # Check for continuing an earlier list item, possibly after an
         # intervening sublist
         if (node.kind == NodeKind.LIST_ITEM and token.endswith(":") and
-            node.args == token[:-1]):
+            node.args == token[:-1] and node.children and
+            isinstance(node.children[-1], WikiNode)):
             # Suffixing a list item prefix with a colon can be used to continue
             # the same item after an intervening sublist.  In this case we
             # just return with the continued list item at the top of the stack.
@@ -974,7 +1036,7 @@ def list_fn(ctx, token):
         _parser_pop(ctx, True)
 
     # If not already in a list, create a new list.
-    node = ctx.stack[-1]
+    node = ctx.parser_stack[-1]
     if node.kind != NodeKind.LIST:
         node = _parser_push(ctx, NodeKind.LIST)
         node.args = token
@@ -1010,8 +1072,15 @@ def tag_fn(ctx, token):
     # Note: <nowiki> and HTML comments have already been handled in
     # preprocessing
 
+    # Do not try to parse HTML inside template arguments.  Many templates
+    # use tag-like constructions as internal markers, and they are not
+    # valid HTML tags.  HTML is only parsed after template processing.
+    if (_parser_have(ctx, NodeKind.TEMPLATE) or
+        _parser_have(ctx, NodeKind.TEMPLATE_ARG)):
+        return text_fn(ctx, token)
+
     # Try to parse it as a start tag
-    m = re.match(r"""<\s*([-a-zA-Z0-9]+)\s*((\b[-a-z0-9]+(=("[^"]*"|"""
+    m = re.match(r"""<\s*([-a-zA-Z0-9]+)\s*((\b[-a-zA-Z0-9]+(=("[^"]*"|"""
                  r"""'[^']*'|[^ \t\n"'`=<>/]*))?\s*)*)(/?)\s*>""", token)
     if m:
         # This is a start tag
@@ -1021,17 +1090,20 @@ def tag_fn(ctx, token):
         name = name.lower()
 
         # If preparsing, only handle template control tags like <noinclude>
-        if ctx.pre_parse and name not in PRE_PARSE_TAGS:
+        if ctx.pre_parse:
             return text_fn(ctx, token)
 
+        # If the tag is <section ...>, ignore it
+        if name == "section":
+            return
+
         # Check for unmatched <nowiki> start tag.  <nowiki> should be handled
         # in preprocessing, but an unmatched start tag may be missed.
         if name == "nowiki":
             if also_end:
-                # XXX mark current link/template/templatearg/parserfn as
-                # nowiki, cause it to be expanded by escaped plaintext
+                text_fn(ctx, MAGIC_NOWIKI_CHAR)
                 return
-            ctx.error("unmatched <nowiki>")
+            ctx.warning("unmatched <nowiki>")
             return text_fn(ctx, token)
 
         # Handle <pre> start tag
@@ -1040,26 +1112,30 @@ def tag_fn(ctx, token):
             parse_attrs(node, attrs)
             if also_end:
                 _parser_pop(ctx, False)
+            else:
+                ctx.pre_parse = True
             return
 
         # Give a warning on unsupported HTML tags.  WikiText limits the set of
         # tags that are allowed.
         if name not in ALLOWED_HTML_TAGS:
-            # Wiktionary seems to use markings like <3> in some
-            # languages.  Treat them as text.  This method of handling
-            # them may need to be reconsidered in the future if
-            # problems arise.
-            if not name.isdigit():
+            if not name.isdigit() and not SILENT_HTML_LIKE:
                 ctx.warning("html tag <{}{}> not allowed in WikiText"
                             "".format(name, "/" if also_end else ""))
-            return text_fn(ctx, token)
+            text_fn(ctx, token)
+            return
 
         # Automatically close parent HTML tags that should be ended by this tag
         # until we have a parent that is not a HTML tag or that is an allowed
         # parent for this node
         permitted_parents = HTML_PERMITTED_PARENTS.get(name, set())
         while True:
-            node = ctx.stack[-1]
+            node = ctx.parser_stack[-1]
+            if node.kind == NodeKind.URL and not node.children:
+                ctx.parser_stack.pop()
+                ctx.parser_stack[-1].children.pop()
+                text_fn(ctx, "[")
+                continue
             if node.kind != NodeKind.HTML:
                 break
             if node.args in permitted_parents:
@@ -1087,29 +1163,35 @@ def tag_fn(ctx, token):
     name = m.group(1)
     name = name.lower()
 
-    # If preparsing, only handle template control tags like <noinclude>
-    if ctx.pre_parse and name not in PRE_PARSE_TAGS:
-        return text_fn(ctx, token)
+    # We should never see </section>
+    if name == "section":
+        ctx.warning("unexpected </section>")
+        return
 
+    # Check for </pre> end tag
     if name == "pre":
         # Handle </pre> end tag
-        node = ctx.stack[-1]
+        ctx.pre_parse = False
+        node = ctx.parser_stack[-1]
         if node.kind != NodeKind.PRE:
-            ctx.error("unexpected </pre>")
+            ctx.warning("unexpected </pre>")
             return text_fn(ctx, token)
         _parser_pop(ctx, False)
         return
 
-    # Give a warning on unsupported HTML tags.  WikiText limits the set of
-    # tags that are allowed.
-    if name not in ALLOWED_HTML_TAGS:
-        ctx.warning("html tag </{}> not allowed in WikiText"
-                    "".format(name))
+    # If preparsing, treat this as plain text
+    if ctx.pre_parse:
         return text_fn(ctx, token)
 
+    # Give a warning on unsupported HTML tags.  WikiText limits the set of
+    # tags that are allowed.
+    if name not in ALLOWED_HTML_TAGS and name != "nowiki":
+        ctx.warning("html tag </{}> not allowed in WikiText"
+                    "".format(name))
+
     # See if we can find the opening tag from the stack
-    for i in range(0, len(ctx.stack)):
-        node = ctx.stack[i]
+    for i in range(0, len(ctx.parser_stack)):
+        node = ctx.parser_stack[i]
         if node.kind == NodeKind.HTML and node.args == name:
             break
     else:
@@ -1120,12 +1202,18 @@ def tag_fn(ctx, token):
             node.args = name
             _parser_pop(ctx, False)
             return
-        ctx.error("no corresponding start tag found for {}".format(token))
+        ctx.warning("no corresponding start tag found for {}".format(token))
+        text_fn(ctx, token)
         return
 
     # Close nodes until we close the corresponding start tag
     while True:
-        node = ctx.stack[-1]
+        node = ctx.parser_stack[-1]
+        if node.kind == NodeKind.URL and not node.children:
+            ctx.parser_stack.pop()
+            ctx.parser_stack[-1].children.pop()
+            text_fn(ctx, "[")
+            continue
         if node.kind == NodeKind.HTML and node.args == name:
             # Found the corresponding start tag.  Close this node and
             # then stop.
@@ -1152,10 +1240,11 @@ def magicword_fn(ctx, token):
 # Regular expression for matching a token in WikiMedia text.  This is used for
 # tokenizing the input.
 token_re = re.compile(r"(?m)^(={2,6})\s*(([^=]|=[^=])+?)\s*(={2,6})\s*$|"
+                      r"'''''|"
                       r"'''|"
                       r"''|"
                       r"[ \t]+\n*|"
-                      r"\n+|"
+                      r"\n|"
                       r"\[|"
                       r"\]|"
                       r"\|\}|"
@@ -1169,10 +1258,11 @@ token_re = re.compile(r"(?m)^(={2,6})\s*(([^=]|=[^=])+?)\s*(={2,6})\s*$|"
                       r"^----+|"
                       r"^[*:;#]+|"
                       r":|"   # sometimes special when not beginning of line
-                      r"""<\s*[-a-zA-Z0-9]+\s*(\b[-a-z0-9]+(=("[^"]*"|"""
-                        r"""'[^']*'|[^ \t\n"'`=<>]*))?\s*)*(/\s*)?>|"""
+                      r"<<[-a-zA-Z0-9/]*>>|"
+                      r"""<\s*[-a-zA-Z0-9]+\s*(\b[-a-zA-Z0-9]+(=("[^<>"]*"|"""
+                        r"""'[^<>']*'|[^ \t\n"'`=<>]*))?\s*)*(/\s*)?>|"""
                       r"<\s*/\s*[-a-zA-Z0-9]+\s*>|"
-                      r"https?://[a-zA-Z0-9.]+(/[^] ]*[^].])?|"
+                      r"https?://[a-zA-Z0-9.]+(/[^][{}<>|\s]*)?|"
                       r"(" +
                       r"|".join(r"\b{}\b".format(x) for x in MAGIC_WORDS) +
                       r")|" +
@@ -1211,55 +1301,119 @@ for x in MAGIC_WORDS:
     tokenops[x] = magicword_fn
 
 
+def bold_follows(parts, i):
+    """Checks if there is a bold (''') in parts after parts[i].  We allow
+    intervening italics ('')."""
+    parts = parts[i + 1:]
+    for p in parts:
+        if not p.startswith("''"):
+            continue
+        if p.startswith("'''"):
+            return True
+    return False
+
+
 def token_iter(ctx, text):
     """Tokenizes MediaWiki page content.  This yields (is_token, text) for
-    each token.  ``is_token`` is False for text and True for other tokens."""
+    each token.  ``is_token`` is False for text and True for other tokens.
+    Wikitext bold and italic are interpreted WITHIN A SINGLE LINE.  It seems
+    impossible to always disambiguate them without looking at what follows
+    on the same line."""
     assert isinstance(text, str)
-    pos = 0
-    for m in re.finditer(token_re, text):
-        start = m.start()
-        if pos != start:
-            yield False, text[pos:start]
-        pos = m.end()
-        token = m.group(0)
-        if token.startswith("=="):
-            yield True, "<" + m.group(1)
-            for x in token_iter(ctx, m.group(2)):
-                yield x
-            yield True, ">" + m.group(4)
-        else:
-            yield True, token
-    if pos != len(text):
-        yield False, text[pos:]
-
-
-def nowiki_sub_fn(m):
-    """This function escapes the contents of a <nowiki> ... </nowiki> pair."""
-    text = m.group(1)
-    text = re.sub(r";", "&semi;", text)
-    text = html.escape(text, quote=True)
-    text = re.sub(r"=", "&equals;", text)
-    text = re.sub(r"\*", "&ast;", text)
-    text = re.sub(r"#", "&num;", text)
-    text = re.sub(r":", "&colon;", text)
-    text = re.sub(r"!", "&excl;", text)
-    text = re.sub(r"\|", "&vert;", text)
-    text = re.sub(r"\[", "&lsqb;", text)
-    text = re.sub(r"\]", "&rsqb;", text)
-    text = re.sub(r"\{", "&lbrace;", text)
-    text = re.sub(r"\}", "&rbrace;", text)
-    text = re.sub(r"\s+", " ", text)
-    return text
-
-
-def preprocess_text(text):
-    # XXX this will be eliminated, needs to be handled differently
-    assert isinstance(text, str)
-    text = re.sub(r"(?si)<\s*nowiki\s*>(.*?)<\s*/\s*nowiki\s*>", nowiki_sub_fn,
-                  text)
-    text = re.sub(r"(?si)<\s*nowiki\s*/\s*>", MAGIC_NOWIKI_CHAR, text)
-    text = re.sub(r"(?s)<!\s*--.*?--\s*>", "", text)
-    return text
+    lines = re.split(r"(\n+)", text)  # Lines and separators
+    parts_re = re.compile(r"(''+)")
+    for line in lines:
+        parts = re.split(parts_re, line)
+        state = 0  # 1=in italic 2=in bold 3=in both
+        for i, part in enumerate(parts):
+            if part.startswith("''"):
+                # This is a bold/italic part.  Scan the rest of the line
+                # to determine how it should be interpreted if there are
+                # more than two apostrophes.
+                if part.startswith("'''''"):
+                    if state == 1:  # in italic
+                        yield True, "''"
+                        yield True, "'''"
+                        part = part[5:]
+                        state = 2
+                    elif state == 2:  # in bold
+                        yield True, "'''"
+                        yield True, "''"
+                        part = part[5:]
+                        state = 1
+                    elif state == 3:  # in both
+                        yield True, "'''"
+                        yield True, "''"
+                        state = 0
+                        part = part[5:]
+                    else:  # in nothing
+                        if bold_follows(parts, i):
+                            yield True, "''"
+                            yield True, "'''"
+                        else:
+                            yield True, "'''"
+                            yield True, "''"
+                        part = part[5:]
+                        state = 3
+                elif part.startswith("'''"):
+                    if state == 1:  # in italic
+                        if bold_follows(parts, i):
+                            yield True, "'''"
+                            part = part[3:]
+                            state = 3
+                        else:
+                            yield True, "''"
+                            part = part[2:]
+                            state = 0
+                    elif state == 2:  # in bold
+                        yield True, "'''"
+                        part = part[3:]
+                        state = 0
+                    elif state == 3:  # in both
+                        yield True, "'''"
+                        part = part[3:]
+                        state = 1
+                    else:  # in nothing
+                        yield True, "'''"
+                        part = part[3:]
+                        state = 2
+                elif part.startswith("''"):
+                    if state == 1:  # in italic
+                        yield True, "''"
+                        part = part[2:]
+                        state = 0
+                    elif state == 2:  # in bold
+                        yield True, "''"
+                        part = part[2:]
+                        state = 3
+                    elif state == 3:  # in both
+                        yield True, "''"
+                        part = part[2:]
+                        state = 2
+                    else:  # in nothing
+                        yield True, "''"
+                        part = part[2:]
+                        state = 1
+                if part:
+                    yield False, part
+                continue
+            # All other parts handled with normal tokenization
+            pos = 0
+            for m in re.finditer(token_re, part):
+                start = m.start()
+                if pos != start:
+                    yield False, part[pos:start]
+                pos = m.end()
+                token = m.group(0)
+                if token.startswith("=="):
+                    yield True, "<" + m.group(1)
+                    for x in token_iter(ctx, m.group(2)):
+                        yield x
+                    yield True, ">" + m.group(4)
+                else:
+                    yield True, token
+            if pos != len(part):
+                yield False, part[pos:]
 
 
 def process_text(ctx, text):
@@ -1267,7 +1421,8 @@ def process_text(ctx, text):
     called recursively (which we do to process tokens inside templates and
     certain other structures)."""
     for is_token, token in token_iter(ctx, text):
-        node = ctx.stack[-1]
+        # print("process_text: token_iter yielded:", is_token, token)
+        node = ctx.parser_stack[-1]
         if not is_token:
             # Process it as normal text.
             text_fn(ctx, token)
@@ -1289,7 +1444,7 @@ def process_text(ctx, text):
                 subtitle_start_fn(ctx, token)
             elif token.startswith(">=="):  # Note: > added by tokenizer
                 subtitle_end_fn(ctx, token)
-            elif token.startswith("<") and len(token):
+            elif token.startswith("<"):
                 tag_fn(ctx, token)
             elif token.startswith("----"):
                 hline_fn(ctx, token)
@@ -1306,35 +1461,39 @@ def process_text(ctx, text):
         ctx.beginning_of_line = token[-1] == "\n"
 
 
-
 def parse_encoded(ctx, text):
-    # XXX comment
+    """Parses the text, which should already have been encoded using magic
+    characters (see Wtp._encode()).  Parses the encoded string and returns
+    the parse tree."""
     assert ctx.title is not None  # ctx.start_page() must have been called
     node = WikiNode(NodeKind.ROOT, 0)
     node.args.append([ctx.title])
     ctx.beginning_of_line = True
     ctx.linenum = 1
     ctx.pre_parse = False
-    ctx.stack = [node]
+    ctx.parser_stack = [node]
     ctx.suppress_special = False
 
+    try:
+        # Process all tokens from the input.
+        process_text(ctx, text)
 
+        # We are at the end of the text.  Keep popping stack until we only have
+        # the root node left.  This is used to finalize processing any nodes
+        # on the stack.
+        while True:
+            node = ctx.parser_stack[-1]
+            if node.kind == NodeKind.ROOT:
+                break
+            _parser_pop(ctx, True)
+        assert len(ctx.parser_stack) == 1
+        # If the last children are strings, merge them to one string.
+        _parser_merge_str_children(ctx)
+        ret = ctx.parser_stack[0]
+    finally:
+        ctx.parser_stack = None
+    return ret
 
-    # Process all tokens from the input.
-    process_text(ctx, text)
-
-    # We are at the end of the text.  Keep popping stack until we only have
-    # the root node left.  This is used to finalize processing any nodes
-    # on the stack.
-    while True:
-        node = ctx.stack[-1]
-        if node.kind == NodeKind.ROOT:
-            break
-        _parser_pop(ctx, True)
-    assert len(ctx.stack) == 1
-    # If the last children are strings, merge them to one string.
-    _parser_merge_str_children(ctx)
-    return ctx.stack[0]
 
 def print_tree(tree, indent=0):
     """Prints the parse tree for debugging purposes.  This does not expand
@@ -1349,9 +1508,3 @@ def print_tree(tree, indent=0):
         print("{}    {}={}".format(" " * indent, k, v))
     for child in tree.children:
         print_tree(child, indent + 2)
-
-
-# Very simple test:
-# data = open("pages/Words/ho/horse.txt", "r").read()
-# tree = parse("horse", data)
-# print_tree(tree, 0)
